import numpy as np

from ... import utils
from ...preprocessing import Preprocessing

from ...model import Model
from ...onnx import ONNXModel


class _MobileNetV2(Model):
    '''
    WARNING: this is not working yet
    '''
    #TODO: make it work

    URL_MODEL = 'https://github.com/onnx/models/blob/master/vision/classification/mobilenet/model/mobilenetv2-7.onnx'
    URL_LABELS = 'https://github.com/onnx/models/blob/master/vision/classification/synset.txt'

    def __init__(self):
        '''
        '''
        super().__init__()

        # Load model
        self.model = ONNXModel(self.URL_MODEL)

        # Load labels from text file
        labels_path = utils.file_from_url(self.URL_LABELS, 'labels')
        self.labels = [
            ' '.join(line.strip().split()[1:])
            for line in open(labels_path)
        ]

    def forward(self, image, return_scores=False):
        scores = self.model(image)[0][0]
        label = self.labels[scores.argmax()]
        if return_scores:
            return label, scores
        else:
            return label

    def get_preprocessing(self):
        return Preprocessing(
            resize=(224, 224),
            dtype='float',
            divide_by_255=True,
            means=[0.485, 0.456, 0.406],
            scales=1 / np.array([0.229, 0.224, 0.225]),
            padding=True,
            batch_dimension=True,
            channels_first=True,
        )



import cv2
import numpy as np
import os

from .model import Model
from . import utils

class Preprocessing(Model):
    def __init__(
        self,
        resize=None,
        dtype='uint8',
        divide_by_255=False,
        means=None,
        scales=None,
        padding=False,
        channels_first=False,
        batch_dimension=False,
        bgr=False,
    ):
        '''
        input:
        - resize: tuple or None
            (width, height) - the new size of image
        - dtype: str on np.dtype
            Data type
        - means: float of list or None
            Substract these values from each channel
        - scales: float of list or None
            Multipy each channel by these values
        - padding: bool
            If True, images will be resized with the same aspect ratio
        - channels_first: bool
            If True, image will be returned in [B]CHW format.
            If False, [B]HWC
        - batch_dimension: bool
            If True, add first dimension of size 1
        - bgr: bool
            If True, change channels to BRG order
            If False, keep the RGB order
        '''
        self.resize = resize
        self.dtype = dtype
        self.divide_by_255 = divide_by_255
        if divide_by_255:
            MSG = 'If dividing image by 255, specify float data type'
            assert 'float' in dtype, MSG
        self.means = means
        self.scales = scales
        self.padding = padding
        self.channels_first = channels_first
        self.batch_dimension = batch_dimension
        self.bgr = bgr

    def forward(self, image, return_shape=False):
        '''
        Preprocess the image.
        input:
        - image: np.ndarray of type uint8 or str
            RGB image
            If str, it will be concerned as image path.
        - return_shape: bool
            If True, will return shape of the original image
        '''
        # Read image
        if isinstance(image, str):
            image = self.read_image(image)
        orig_shape = image.shape

        # Convert colors
        if self.bgr:
            image = image[:, :, ::-1]

        # Resize image
        if self.resize is not None:
            image = self.resize_image(image, self.resize, self.padding)

        # Change datatype
        assert str(image.dtype) == 'uint8'
        image = image.astype(self.dtype)

        # Divide by 255
        if self.divide_by_255:
            image = image / self.divide_by_255

        # Shift and scale
        if self.means is not None:
            image = image - np.array(self.means)[None, None]
        if self.scales is not None:
            image = image * np.array(self.scales)[None, None]

        # Change shape
        if self.channels_first:
            image = image.transpose([2, 0, 1])
        if self.batch_dimension:
            image = image[None]

        if return_shape:
            return image.copy(), orig_shape
        else:
            return image.copy()

    @staticmethod
    def read_image(path):
        ''' Read image from file or url '''
        # Download image if path is url
        is_url = utils.is_url(path)
        if is_url:
            path = utils.file_from_url(path, 'temp')
        # Read image
        image = cv2.imread(path)
        # Delete temporary file
        if is_url:
            os.remove(path)
        # Throw exception
        if image is None:
            raise BaseException('Cannot read ' + image_path)
        # Convert from BGR to RGB
        image = image[:, :, ::-1]
        return image

    @staticmethod
    def resize_image(image, resize, padding=False):
        ''' Resize image
        input:
        - image
        - resize: tuple
            (width, height) - the new size
        - padding: bool
            If True, image will be resized with the same aspect ratio
        '''
        if not padding:
            image = cv2.resize(image, resize)
        else:
            # Resize saving the aspect ratio
            ratio_0 = image.shape[1] / resize[0]
            ratio_1 = image.shape[0] / resize[1]
            ratio = max(ratio_0, ratio_1)
            new_size = (
                int(image.shape[1] / ratio),
                int(image.shape[0] / ratio)
            )
            image = cv2.resize(image, new_size)
            # Pad with zeros
            padding = np.zeros([resize[1], resize[0], 3],
                               dtype=image.dtype)
            start_0 = (padding.shape[0] - image.shape[0]) // 2
            start_1 = (padding.shape[1] - image.shape[1]) // 2
            padding[
                start_0: start_0 + image.shape[0],
                start_1: start_1 + image.shape[1],
            ] = image
            image = padding.copy()
        return image






















from ... import utils
from ...preprocessing import Preprocessing

from ...model import Model
from ...edgetpu import EdgeTPUModel
from ...output import DetectionBox

class SSDMobileNet(Model):
    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_{}_coco_quant_postprocess.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_{}_coco_quant_postprocess_edgetpu.tflite'
    URL_LABELS = 'https://github.com/google-coral/edgetpu/raw/master/test_data/coco_labels.txt'

    def __init__(
        self,
        device=None,
        version='v2',
        threshold=0.5
    ):
        '''
        input:
        - device: str or None
            Set ":0" to use the first EdgeTPU device.
            Set ":1" to use the second EdgeTPU device.
            Same for other devices if they are present.
            Leave None to use CPU
        - version: str
            Either "v1" or "v2"
        - threshold: float
            Detection threshold
        '''
        super().__init__()

        self.threshold = threshold

        # Load model
        if device is None:
            model_path = self.URL_CPU.format(version)
        else:
            model_path = self.URL_TPU.format(version)
        self.model = EdgeTPUModel(model_path, device)

        # Load labels from text file
        labels_path = utils.file_from_url(self.URL_LABELS, 'labels_google')
        self.labels = {
            int(line.split()[0]): line.strip().split()[1]
            for line in open(labels_path)
        }

    def forward(self, image):
        boxes, classes, scores, num_detections = self.model(image)
        # Parse output
        out_boxes = []
        for i in range(len(boxes[0])):
            if scores[0, i] < self.threshold:
                continue
            x_1, y_1, x_2, y_2 = boxes[0, i]
            label = self.labels[int(classes[0, i])]
            score = scores[0, i]
            out_boxes.append(
                DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        return out_boxes

    def get_preprocessing(self):
        return Preprocessing(
            resize=(300, 300),
            dtype='uint8',
            batch_dimension=True,
        )


class SSDMobileNetFace(Model):
    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_v2_face_quant_postprocess.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/ssd_mobilenet_v2_face_quant_postprocess_edgetpu.tflite'

    def __init__(
        self,
        device=None,
        threshold=0.5
    ):
        '''
        input:
        - device: str or None
            Set ":0" to use the first EdgeTPU device.
            Set ":1" to use the second EdgeTPU device.
            Same for other devices if they are present.
            Leave None to use CPU
        - threshold: float
            Detection threshold
        '''
        super().__init__()

        self.threshold = threshold

        # Load model
        if device is None:
            model_path = self.URL_CPU
        else:
            model_path = self.URL_TPU
        self.model = EdgeTPUModel(model_path, device)

    def forward(self, image):
        boxes, _, scores, num_detections = self.model(image)
        # Parse output
        out_boxes = []
        for i in range(len(boxes[0])):
            if scores[0, i] < self.threshold:
                continue
            x_1, y_1, x_2, y_2 = boxes[0, i]
            label = 'face'
            score = scores[0, i]
            out_boxes.append(
                DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        return out_boxes

    def get_preprocessing(self):
        return Preprocessing(
            resize=(320, 320),
            dtype='uint8',
            batch_dimension=True,
        )



from ... import utils
from ...preprocessing import Preprocessing

from ...model import Model
from ...onnx import ONNXModel
from ...output import DetectionBox

class SSDMobileNetV1(Model):
    URL_MODEL = 'https://github.com/onnx/models/raw/master/vision/object_detection_segmentation/ssd-mobilenetv1/model/ssd_mobilenet_v1_10.onnx'
    URL_LABELS = 'https://github.com/amikelive/coco-labels/raw/master/coco-labels-paper.txt'

    def __init__(self):
        '''
        '''
        super().__init__()

        # Load model
        self.model = ONNXModel(self.URL_MODEL)

        # Load labels from text file
        labels_path = utils.file_from_url(self.URL_LABELS, 'labels')
        self.labels = [
            line.strip()
            for line in open(labels_path)
        ]

    def forward(self, image):
        boxes, classes, scores, num_detections = self.model(image)
        # Parse output
        classes = classes - 1
        out_boxes = []
        for i in range(int(num_detections[0])):
            x_1, y_1, x_2, y_2 = boxes[0, i]
            label = self.labels[int(classes[0, i])]
            score = scores[0, i]
            out_boxes.append(
                DetectionBox(x_1, y_1, x_2, y_2, label, score)
            )
        return out_boxes

    def get_preprocessing(self):
        return Preprocessing(
            resize=(224, 224),
            dtype='uint8',
            batch_dimension=True,
        )



import platform
import time

from .model import Model
from . import utils

EDGETPU_SHARED_LIB = {
  'Linux': 'libedgetpu.so.1',
  'Darwin': 'libedgetpu.1.dylib',
  'Windows': 'edgetpu.dll'
}[platform.system()]

class EdgeTPUModel(Model):
    def __init__(
        self,
        model_path,
        device=None
    ):
        '''
        input:
        - model_path: str
            url or path to the tflite model
        - device: str or None
            Set ":0" to use the first EdgeTPU device.
            Set ":1" to use the second EdgeTPU device.
            Same for other devices if they are present.
            Leave None to use CPU
        '''
        super().__init__()
        # Download file from internet
        if utils.is_url(model_path):
            model_path = utils.file_from_url(model_path, 'models')
        # Create interpreter
        self.interpreter = self.make_interpreter(model_path, device)
        self.interpreter.allocate_tensors()

    def forward(self, *inputs, return_time=False):
        '''
        input:
        - *inputs: list of arguments
            Input numpy arrays
        - return_time: bool
            If True, will return inference time
        '''
        assert len(inputs) == self.n_inputs
        start = time.time()
        # Put input tensors into model
        for i in range(self.n_inputs):
            tensor = self.input_tensor(i)
            tensor[:, :, :, :] = inputs[i]
            del tensor
        before_invoke = time.time()
        # Call model
        self.interpreter.invoke()
        after_invoke = time.time()
        # Get results from the model
        results = [self.output_tensor(i) for i in range(self.n_outputs)]
        end = time.time()
        times = {
            'assign': before_invoke - start,
            'invoke': after_invoke - before_invoke,
            'getres': end - after_invoke,
        }
        if return_time:
            return results, times
        else:
            return results

    def get_input_details(self):
        return [
            {
                'name': inp['name'],
                'shape': inp['shape'],
                'dtype': str(inp['dtype']),
            }
            for inp in self.interpreter.get_input_details()
        ]

    def get_output_details(self):
        return [
            {
                'name': inp['name'],
                'shape': inp['shape'],
                'dtype': str(inp['dtype']),
            }
            for inp in self.interpreter.get_output_details()
        ]

    @staticmethod
    def make_interpreter(model_file, device=None):
        import tflite_runtime.interpreter as tflite
        if device is not None:
            return tflite.Interpreter(
                model_path=model_file,
                experimental_delegates=[
                    tflite.load_delegate(
                        EDGETPU_SHARED_LIB,
                        {'device': device})
                ])
        else:
            return tflite.Interpreter(
                model_path=model_file)

    def input_tensor(self, i=0):
        '''
        Returns input tensor view as function returning numpy array
        '''
        tensor_index = self.interpreter.get_input_details()[i]['index']
        return self.interpreter.tensor(tensor_index)()

    def output_tensor(self, i=0):
        """Returns output tensor view."""
        tensor_index = self.interpreter.get_output_details()[i]['index']
        tensor = self.interpreter.get_tensor(tensor_index)
        return tensor

    @property
    def n_inputs(self):
        ''' number of input tensors '''
        return len(self.interpreter.get_input_details())

    @property
    def n_outputs(self):
        ''' number of output tensors '''
        return len(self.interpreter.get_output_details())




import time

from .model import Model
from . import utils

class ONNXModel(Model):
    def __init__(
        self,
        model_path,
    ):
        '''
        input:
        - model_path: str
            url or path to the .onnx model
        '''
        super().__init__()
        # Download file from internet
        if utils.is_url(model_path):
            model_path = utils.file_from_url(model_path, 'models')
        # Load model and create inference session
        import onnxruntime as rt
        self.sess = rt.InferenceSession(model_path)

    def forward(self, *inputs, return_time=False):
        '''
        input:
        - *inputs: list of arguments
            Input numpy arrays
        - return_time: bool
            If True, will return inference time
        '''
        assert len(inputs) == len(self.get_input_details())
        # List output names
        outputs = [
            info['name']
            for info in self.get_output_details()
        ]
        # Convert input to a dict
        inputs = {
            info['name']: inp
            for (info, inp) in zip(self.get_input_details(), inputs)
        }
        # Run network and measure time
        start = time.time()
        results = self.sess.run(outputs, inputs)
        end = time.time()
        times = {
            'invoke': end - start,
        }
        if return_time:
            return results, times
        else:
            return results

    def get_input_details(self):
        return [
            {
                'name': info.name,
                'shape': info.shape,
                'dtype': info.type,
            }
            for info in self.sess.get_inputs()
        ]

    def get_output_details(self):
        return [
            {
                'name': info.name,
                'shape': info.shape,
                'dtype': info.type,
            }
            for info in self.sess.get_outputs()
        ]



import urllib.request
import os
import pathlib
import getpass

PACKAGE_NAME = 'nnio'

def is_url(s):
    '''
    Check if input string is url or not
    '''
    return s.startswith('http://') or s.startswith('https://')

def file_from_url(url, category='other'):
    '''
    Downloads file to "/home/$USER/.cache/nnio/" if it does not exist already.
    Returns path to the file
    '''
    # Get base path for file
    base_path = os.path.join(
        '/home',
        getpass.getuser(),
        '.cache',
        PACKAGE_NAME,
        category,
    )
    # Create path if not exists
    if not os.path.exists(base_path):
        pathlib.Path(base_path).mkdir(parents=True, exist_ok=True)
    # Get file path
    file_name = url.split('/')[-1]
    file_path = os.path.join(
        base_path,
        file_name
    )
    # Download file from the url
    if not os.path.exists(file_path):
        print('Downloading file from: {}'.format(url))
        urllib.request.urlretrieve(url, file_path)
    else:
        print('Using cached file: {}'.format(file_path))

    return file_path



import abc


class Model(abc.ABC):
    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        '''
        This function is called when the model is called.
        Override this function with your model's behavior
        '''

    def get_preprocessing(self):
        '''
        Overriding this function is recommended.
        It should return the callable object for image preprocessing
        For example:
        return nnio.Preprocessing(
            resize=(224, 224),
            dtype='uint8',
            padding=True,
            batch_dimension=True,
        )
        '''

    def get_input_details(self):
        '''
        Overriding this function is recommended.
        It should return human-readable input details.
        For example:
        [
            {
                'name': 'input_image'
                'shape': [1, 224, 224, 3]
                'dtype': 'uint8'
            }
        ]
        '''

    def get_output_details(self):
        '''
        Overriding this function is recommended.
        It should return human-readable output details.
        For example:
        [
            {
                'name': 'detection_boxes'
                'shape': [10, 14, 14, 4]
                'type': 'float'
            }
            {
                'name': 'detection_classes'
                'shape': [10]
                'type': 'float'
            }
        ]
        '''



import nnio.zoo

# Base model class
from .model import Model

# Models for specific backends
from .edgetpu import EdgeTPUModel
from .openvino import OpenVINOModel
from .onnx import ONNXModel

# Preprocessing class
from .preprocessing import Preprocessing



from ... import utils
from ...preprocessing import Preprocessing

from ...model import Model
from ...edgetpu import EdgeTPUModel


class MobileNet(Model):
    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/mobilenet_{}_1.0_224_quant.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/mobilenet_{}_1.0_224_quant_edgetpu.tflite'
    URL_LABELS = 'https://github.com/google-coral/edgetpu/raw/master/test_data/imagenet_labels.txt'

    def __init__(self, device=None, version='v2'):
        '''
        input:
        - device: str or None
            Set ":0" to use the first EdgeTPU device.
            Set ":1" to use the second EdgeTPU device.
            Same for other devices if they are present.
            Leave None to use CPU
        - version: str
            Either "v1" or "v2"
        '''
        super().__init__()

        # Load model
        if device is None:
            model_path = self.URL_CPU.format(version)
        else:
            model_path = self.URL_TPU.format(version)
        self.model = EdgeTPUModel(model_path, device)

        # Load labels from text file
        labels_path = utils.file_from_url(self.URL_LABELS, 'labels')
        self.labels = [
            ' '.join(line.strip().split()[1:])
            for line in open(labels_path)
        ]

    def forward(self, image, return_scores=False):
        scores = self.model(image)[0][0]
        label = self.labels[scores.argmax()]
        if return_scores:
            return label, scores
        else:
            return label

    def get_preprocessing(self):
        return Preprocessing(
            resize=(224, 224),
            dtype='uint8',
            padding=True,
            batch_dimension=True,
        )



import cv2


class DetectionBox:
    def __init__(
        self,
        x_1,
        y_1,
        x_2,
        y_2,
        label=None,
        score=1.0,
    ):
        '''
        inputs:
        - x_1: float in range [0, 1]
            Relative x coordinate of top-left corner
        - y_1: float in range [0, 1]
            Relative y coordinate of top-left corner
        - x_2: float in range [0, 1]
            Relative x coordinate of bottom-right corner
        - y_2: float in range [0, 1]
            Relative y coordinate of bottom-right corner
        - label: str or None
            Class label of the detected object
        - score: float
            Detection score
        '''
        self.x_1 = x_1
        self.x_2 = x_2
        self.y_1 = y_1
        self.y_2 = y_2
        self.label = label
        self.score = score

    def draw(
        self,
        image,
        color=(255,0,0),
        stroke_width=2,
        text_color=(255,0,0),
        text_width=2,
    ):
        '''
        Draws the detection box on image

        inputs:
        - image: numpy array
        -- Drawing parameters

        output:
        -- Modified image
        '''
        # Box corners:
        start_point = (
            int(image.shape[1] * self.y_1),
            int(image.shape[0] * self.x_1),
        )
        end_point = (
            int(image.shape[1] * self.y_2),
            int(image.shape[0] * self.x_2),
        )
        # Draw rectangle
        image = cv2.rectangle(
            image,
            start_point,
            end_point,
            color, stroke_width)
        # Draw text
        if self.label is not None:
            image = cv2.putText(
                image,
                self.label,
                start_point,
                cv2.FONT_HERSHEY_SIMPLEX,
                1, text_color, text_width, cv2.LINE_AA
            )
        return image

    def __str__(self):
        template = 'DetectionBox(x_1={}, y_1={}, x_2={}, y_2={}, label={}, score={})'
        s = template.format(
            self.x_1, self.y_1,
            self.x_2, self.y_2,
            self.label,
            self.score)
        return s




import pathlib
from distutils.core import setup, Extension

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "readme.md").read_text()

setup(
    name='nnio',
    version='0.1.4',
    description='Simple neural network inference on devices',
    long_description='''
    See more details at the project's homepage: https://github.com/FastSense/nnio
    ''',
    author='FastSense',
    author_email='ruslan.baynazarov@fastsense.tech',
    url='https://github.com/FastSense/nnio',
    packages=['nnio', 'nnio.zoo', 'nnio.zoo.edgetpu', 'nnio.zoo.openvino', 'nnio.zoo.onnx'],
    install_requires=[
        'opencv-python'
    ]
)



import time

from .model import Model
from . import utils

class OpenVINOModel(Model):
    def __init__(
        self,
        model_bin,
        model_xml,
        device='CPU',
    ):
        '''
        input:
        - model_bin: str
            url or path to the openvino binary model file
        - model_xml: str
            url or path to the openvino xml model file
        - device: str
            Choose Intel device:
            "CPU", "GPU", "MYRIAD"
        '''
        super().__init__()

        # Download files from internet
        if utils.is_url(model_bin):
            model_bin = utils.file_from_url(model_bin, 'models')
        if utils.is_url(model_xml):
            model_xml = utils.file_from_url(model_xml, 'models')

        # Create interpreter
        from openvino.inference_engine import IECore
        ie = IECore()
        self.net = ie.read_network(model_xml, model_bin)
        self.net = ie.load_network(self.net, device)

    def forward(self, inputs, return_time=False):
        '''
        input:
        - inputs: numpy array
            Input data
        - return_time: bool
            If True, will return inference time
        '''
        # Find name of the input to the model
        input_name = list(self.net.input_info.keys())[0]
        # Call model
        start = time.time()
        out = self.net.infer({input_name: inputs})
        end = time.time()
        # Process output a little
        if len(out.keys()) == 1:
            out = out[list(out.keys())[0]]
        times = {
            'invoke': end - start,
        }
        if return_time:
            return out, times
        else:
            return out
